"""
A set of objects that are suited to being the basis
of font development tools. This works on UFO files.
"""
from __future__ import absolute_import

try:
    from ._version import __version__
except ImportError:
    try:
        from setuptools_scm import get_version
        __version__ = get_version()
    except ImportError:
        __version__ = 'unknown'

from defcon.errors import DefconError

from defcon.objects.font import Font
from defcon.objects.layerSet import LayerSet
from defcon.objects.layer import Layer
from defcon.objects.glyph import Glyph, addRepresentationFactory, removeRepresentationFactory
from defcon.objects.contour import Contour
from defcon.objects.point import Point
from defcon.objects.component import Component
from defcon.objects.anchor import Anchor
from defcon.objects.image import Image
from defcon.objects.info import Info
from defcon.objects.groups import Groups
from defcon.objects.kerning import Kerning
from defcon.objects.features import Features
from defcon.objects.lib import Lib
from defcon.objects.uniData import UnicodeData
from defcon.objects.color import Color
from defcon.objects.guideline import Guideline
from defcon.objects.layoutEngine import LayoutEngine

def registerRepresentationFactory(cls, name, factory, destructiveNotifications=None):
    """
    Register **factory** as a representation factory
    for all instances of **cls** (a :class:`defcon.objects.base.BaseObject`)
    subclass under **name**.
    """
    if destructiveNotifications is None:
        destructiveNotifications = [cls.changeNotificationName]
    destructiveNotifications = set(destructiveNotifications)
    cls.representationFactories[name] = dict(factory=factory, destructiveNotifications=destructiveNotifications)

def unregisterRepresentationFactory(cls, name):
    """
    Unregister the representation factory stored under
    **name** in all instances of **cls**.
    """
    del cls.representationFactories[name]



import unittest
import os
import shutil
import fs
import fs.copy
import fs.path
from defcon import Font
from defcon.test.testTools import (
    getTestFontPath, getTestFontCopyPath, makeTestFontCopy,
    openTestFontAsFileSystem, closeTestFontAsFileSystem,
    tearDownTestFontCopy)
from fontTools.ufoLib import UFOReader


class DataSetTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def tearDown(self):
        path = getTestFontCopyPath()
        if os.path.exists(path):
            shutil.rmtree(path)

    def test_read(self):
        path = getTestFontPath()
        font = Font(path)
        fileNames = [
            "com.typesupply.defcon.test.directory/file 1.txt",
            "com.typesupply.defcon.test.directory/sub directory/file 2.txt",
            "com.typesupply.defcon.test.file"]
        for i, fileName in enumerate(sorted(font.data.fileNames)):
            if True in [j.startswith(".") for j in fileName.split(os.sep)]:
                continue
            self.assertEqual(fileName, fileNames[i])
        self.assertEqual(
            font.data["com.typesupply.defcon.test.directory/file 1.txt"],
            b"This is file 1.")
        self.assertEqual(
            font.data["com.typesupply.defcon.test.directory/sub directory/file 2.txt"],
            b"This is file 2.")
        self.assertEqual(
            font.data["com.typesupply.defcon.test.file"],
            b"This is a top level test file.")

    def test_write(self):
        path = makeTestFontCopy()
        with Font(path) as font:
            font.data["com.typesupply.defcon.test.newdirectory/file.txt"] = b"hello."
            del font.data["com.typesupply.defcon.test.directory/sub directory/file 2.txt"]
            font.save()
            p = os.path.join(path, "data",
                             "com.typesupply.defcon.test.newdirectory", "file.txt")
            self.assertTrue(os.path.exists(p))
            with open(p, "r") as f:
                t = f.read()
            self.assertEqual(t, "hello.")
            p = os.path.join(path, "data",
                             "com.typesupply.defcon.test.directory",
                             "sub directory", "file 2.txt")
            self.assertFalse(os.path.exists(p))
        tearDownTestFontCopy()

    def test_save_as(self):
        path = getTestFontPath()
        with Font(path) as font:
            saveAsPath = getTestFontCopyPath(path)
            font.save(saveAsPath)
            dataDirectory = os.path.join(saveAsPath, "data")
            self.assertTrue(os.path.exists(dataDirectory))
            self.assertTrue(os.path.exists(os.path.join(
                dataDirectory,
                os.path.join("com.typesupply.defcon.test.directory",
                             "file 1.txt"))))
            self.assertTrue(os.path.exists(os.path.join(
                dataDirectory,
                os.path.join("com.typesupply.defcon.test.directory",
                             "sub directory", "file 2.txt"))))
            self.assertTrue(os.path.exists(os.path.join(
                dataDirectory,
                "com.typesupply.defcon.test.file")))
        tearDownTestFontCopy(saveAsPath)

    def test_testForExternalChanges_remove_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                del font.data["com.typesupply.defcon.test.file"]
                with UFOReader(path) as reader:
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_add_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                font.data["com.typesupply.defcon.test.file2"] = "blah"
                with UFOReader(path) as reader:
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_modify_in_memory_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                with UFOReader(path) as reader:
                    font.data["com.typesupply.defcon.test.file"] = "blah"
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 ([], [], []))
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_remove_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                # image = font.data["com.typesupply.defcon.test.file"]
                font.data["com.typesupply.defcon.test.file"]
                fileSystem = openTestFontAsFileSystem(font.path)
                fileSystem.remove(fs.path.join("data",
                                       "com.typesupply.defcon.test.file"))
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 ([], [], ["com.typesupply.defcon.test.file"]))
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_add_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                fileSystem = openTestFontAsFileSystem(font.path)
                source = fs.path.join("data", "com.typesupply.defcon.test.file")
                dest = fs.path.join("data", "com.typesupply.defcon.test.file2")
                fileSystem.copy(source, dest)
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 ([], ["com.typesupply.defcon.test.file2"], []))
            tearDownTestFontCopy(font.path)

    def test_testForExternalChanges_modify_on_disk_and_scan(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                # d = font.data["com.typesupply.defcon.test.file"]
                font.data["com.typesupply.defcon.test.file"]
                fileSystem = openTestFontAsFileSystem(font.path)
                filePath = fs.path.join("data",
                                        "com.typesupply.defcon.test.file")
                fileSystem.writebytes(filePath, b"blah")
                closeTestFontAsFileSystem(fileSystem, font.path)
                with UFOReader(path) as reader:
                    self.assertEqual(font.data.testForExternalChanges(reader),
                                 (["com.typesupply.defcon.test.file"], [], []))
            tearDownTestFontCopy(font.path)

    def test_reload_data(self):
        path = makeTestFontCopy()
        with Font(path) as font:
            # d = font.data["com.typesupply.defcon.test.file"]
            font.data["com.typesupply.defcon.test.file"]
            filePath = os.path.join(path, "data",
                                    "com.typesupply.defcon.test.file")
            newData = b"blah"
            f = open(filePath, "wb")
            f.write(newData)
            f.close()
            font.data.reloadData(["com.typesupply.defcon.test.file"])
            data = font.data["com.typesupply.defcon.test.file"]
            self.assertEqual(data, newData)
        tearDownTestFontCopy()


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from fontTools.ufoLib.kerning import lookupKerningValue
from defcon.objects.base import BaseDictObject


class Kerning(BaseDictObject):

    """
    This object contains all of the kerning pairs in a font.

    **This object posts the following notifications:**

    - Kerning.Changed
    - Kerning.BeginUndo
    - Kerning.EndUndo
    - Kerning.BeginRedo
    - Kerning.EndRedo
    - Kerning.PairSet
    - Kerning.PairDeleted
    - Kerning.Cleared
    - Kerning.Updated

    This object behaves like a dict. For example, to get a list of all kerning pairs::

        pairs = kerning.keys()

    To get all pairs including the values::

        for (left, right), value in kerning.items():

    To get the value for a particular pair::

        value = kerning["a", "b"]

    To set the value for a particular pair::

        kerning["a", "b"] = 100

    And so on.

    **Note:** This object is not very smart in the way it handles zero values,
    exceptions, etc. This may change in the future.
    """

    changeNotificationName = "Kerning.Changed"
    beginUndoNotificationName = "Kerning.BeginUndo"
    endUndoNotificationName = "Kerning.EndUndo"
    beginRedoNotificationName = "Kerning.BeginRedo"
    endRedoNotificationName = "Kerning.EndRedo"
    setItemNotificationName = "Kerning.PairSet"
    deleteItemNotificationName = "Kerning.PairDeleted"
    clearNotificationName = "Kerning.Cleared"
    updateNotificationName = "Kerning.Updated"
    representationFactories = {}

    def __init__(self, font=None):
        self._font = None
        if font is not None:
            self._font = weakref.ref(font)
        super(Kerning, self).__init__()
        self.beginSelfNotificationObservation()

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # -------------
    # Pair Handling
    # -------------

    def get(self, pair, default=0):
        return super(Kerning, self).get(pair, default)

    def find(self, pair, default=0):
        """
        This will find the value for **pair** even if
        **pair** is not specifically defined. For example:
        You have a group named `public.kern1.A` with
        the contents `["A", "Aacute"]` and you have a
        group named `public.kern2.C` with the contents
        `["C", "Ccedilla"]`. The only defined kerning is
        `("public.kern1.A", public.kern2.C) = 100`.
        If you use this method to find the value for
        `("A", "Ccedilla")` you will get `100`.
        """
        from defcon.errors import DefconError
        font = self.font
        if font is None:
            raise DefconError("The find method requires a parent font.")
        glyphToFirstGroup = font.groups.getRepresentation("defcon.groups.kerningGlyphToSide1Group")
        glyphToSecondGroup = font.groups.getRepresentation("defcon.groups.kerningGlyphToSide2Group")
        return lookupKerningValue(pair, self, font.groups, fallback=default, glyphToFirstGroup=glyphToFirstGroup, glyphToSecondGroup=glyphToSecondGroup)

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Kerning, self).endSelfNotificationObservation()
        self._font = None


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from fontTools.pens.pointPen import AbstractPointPen
from warnings import warn


class TransformPointPen(AbstractPointPen):

    """PointPen that transforms all coordinates, and passes them to another
    PointPen. It also transforms the transformation given to addComponent().
    """

    def __init__(self, outPen, transformation):
        if not hasattr(transformation, "transformPoint"):
            from fontTools.misc.transform import Transform
            transformation = Transform(*transformation)
        self._transformation = transformation
        self._transformPoint = transformation.transformPoint
        self._outPen = outPen
        self._stack = []

    def beginPath(self, identifier=None):
        self._outPen.beginPath(identifier=identifier)

    def endPath(self):
        self._outPen.endPath()

    def addPoint(self, pt, segmentType=None, smooth=False, name=None, **kwargs):
        pt = self._transformPoint(pt)
        self._outPen.addPoint(pt, segmentType, smooth, name, **kwargs)

    def addComponent(self, glyphName, transformation, identifier=None):
        transformation = self._transformation.transform(transformation)
        try:
            self._outPen.addComponent(glyphName, transformation, identifier)
        except TypeError:
            self._outPen.addComponent(glyphName, transformation)
            warn("The addComponent method needs an identifier kwarg. The component's identifier value has been discarded.", DeprecationWarning)



import unittest
from defcon import DefconError


class DefconErrorTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_DefconError(self):
        with self.assertRaisesRegex(DefconError, "foobar"):
            raise DefconError("foobar")

    def test_report(self):
        defconError = DefconError("foo")
        defconError.report = "bar"
        self.assertEqual(defconError.report, "bar")

if __name__ == "__main__":
    unittest.main()



from fontTools.pens.pointPen import AbstractPointPen

class GlyphObjectPointPen(AbstractPointPen):

    def __init__(self, glyph):
        self._glyph = glyph
        self._contour = None
        self.skipConflictingIdentifiers = False

    def beginPath(self, identifier=None, **kwargs):
        self._contour = self._glyph.instantiateContour()
        self._contour.disableNotifications()
        if identifier is not None:
            if self.skipConflictingIdentifiers and identifier in self._glyph.identifiers:
                pass
            else:
                self._contour.identifier = identifier

    def endPath(self):
        self._contour.dirty = False
        self._glyph.appendContour(self._contour)
        self._contour.enableNotifications()
        self._contour = None

    def addPoint(self, pt, segmentType=None, smooth=False, name=None, identifier=None, **kwargs):
        if self.skipConflictingIdentifiers and identifier in self._glyph.identifiers:
            identifier = None
        self._contour.addPoint(pt, segmentType, smooth, name, identifier=identifier)

    def addComponent(self, baseGlyphName, transformation, identifier=None, **kwargs):
        if self.skipConflictingIdentifiers and identifier in self._glyph.identifiers:
            identifier = None
        component = self._glyph.instantiateComponent()
        component.baseGlyph = baseGlyphName
        component.transformation = transformation
        component.identifier = identifier
        self._glyph.appendComponent(component)


class GlyphObjectLoadingPointPen(GlyphObjectPointPen):

    def __init__(self, glyph):
        super(GlyphObjectLoadingPointPen, self).__init__(glyph)
        self._contours = glyph._shallowLoadedContours

    def beginPath(self, identifier=None, **kwargs):
        contour = dict(points=[])
        if identifier is not None and self.skipConflictingIdentifiers and identifier in self._glyph.identifiers:
            identifier = None
        if identifier is not None:
            if identifier in self._glyph.identifiers:
                raise DefconError("The contour identifier (%s) is already used." % identifier)
            # FIXME: we should do self._glyph.identifiers.add(identifier)
            # otherwise the shallow contours could define the same identifier multiple times
            # or even between shallow loading and real loading something else could
            # take the identifier. The check above is pretty much worthless
            # without storing the identifier.
            contour["identifier"] = identifier
        self._contours.append(contour)

    def endPath(self):
        pass

    def addPoint(self, pt, segmentType=None, smooth=False, name=None, identifier=None, **kwargs):
        args = (pt,)
        kwargs = dict(
            segmentType=segmentType,
            smooth=smooth,
            name=name
        )
        if identifier is not None and self.skipConflictingIdentifiers and identifier in self._glyph.identifiers:
            identifier = None
        if identifier is not None:
            if identifier in self._glyph.identifiers:
                raise DefconError("The contour identifier (%s) is already used." % identifier)
            # FIXME: we should do self._glyph.identifiers.add(identifier)
            # otherwise the shallow contours could define the same identifier multiple times
            # or even between shallow loading and real loading something else could
            # take the identifier. The check above is pretty much worthless
            # without storing the identifier.
            kwargs["identifier"] = identifier
        self._contours[-1]["points"].append((args, kwargs))



class DefconError(Exception):

    _report = None

    def _set_report(self, value):
        self._report = value

    def _get_report(self):
        return self._report

    report = property(_get_report, _set_report)



import random

characters = list("0123456789")
characters += list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
characters += list("abcdefghijklmnopqrstuvwxyz")
identifierLength = 10
identifierRange = range(identifierLength)

def makeRandomIdentifier(existing, recursionDepth=0):
    if recursionDepth >= 50:
        raise NotImplementedError("Failed 50 times in a row to create a unique id. Sorry.")
    identifier = []
    for i in identifierRange:
        c = random.choice(characters)
        identifier.append(c)
    identifier = "".join(identifier)
    if identifier in existing:
        return makeRandomIdentifier(existing, recursionDepth+1)
    else:
        return identifier



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseDictObject


class Lib(BaseDictObject):

    """
    This object contains arbitrary data.

    **This object posts the following notifications:**

    - Lib.Changed
    - Lib.BeginUndo
    - Lib.EndUndo
    - Lib.BeginRedo
    - Lib.EndRedo
    - Lib.ItemSet
    - Lib.ItemDeleted
    - Lib.Cleared
    - Lib.Updated

    This object behaves like a dict. For example, to get a particular
    item from the lib::

        data = lib["com.typesupply.someApplication.blah"]

    To set the glyph list for a particular group name::

        lib["com.typesupply.someApplication.blah"] = 123

    And so on.

    **Note 1:** It is best to keep the data below the top level as shallow
    as possible. Changes below the top level will go unnoticed by the defcon
    change notification system. These changes will be saved the next time you
    save the font, however.

    **Note 2:** The keys used for storing data in the lib should follow the
    reverse domain naming convention detailed in the
    `UFO specification <http://unifiedfontobject.org/filestructure/lib.html>`_.
    """

    changeNotificationName = "Lib.Changed"
    beginUndoNotificationName = "Lib.BeginUndo"
    endUndoNotificationName = "Lib.EndUndo"
    beginRedoNotificationName = "Lib.BeginRedo"
    endRedoNotificationName = "Lib.EndRedo"
    setItemNotificationName = "Lib.ItemSet"
    deleteItemNotificationName = "Lib.ItemDeleted"
    clearNotificationName = "Lib.Cleared"
    updateNotificationName = "Lib.Updated"
    representationFactories = {}

    def __init__(self, font=None, layer=None, glyph=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        if font is not None:
            self.font = font
        if layer is not None:
            self.layer = layer
        if glyph is not None:
            self.glyph = glyph
        super(Lib, self).__init__()
        self.beginSelfNotificationObservation()

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        if self._font is not None:
            return self.font
        elif self._layer is not None:
            return self.layer
        elif self._glyph is not None:
            return self.glyph
        return None

    def _get_font(self):
        font = None
        if self._font is None:
            layerSet = self.layerSet
            if layerSet is not None:
                font = layerSet.font
                if font is not None:
                    self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    def _set_font(self, font):
        assert self._font is None
        assert self._layer is None
        assert self._glyph is None
        if font is not None:
            font = weakref.ref(font)
        self._font = font

    font = property(_get_font, _set_font, doc="The :class:`Font` that this object belongs to. This should not be set externally.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            layer = self.layer
            if layer is not None:
                layerSet = layer.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this object belongs to (if it isn't a font lib).")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    def _set_layer(self, layer):
        assert self._font is None
        assert self._layer is None
        assert self._glyph is None
        if layer is not None:
            layer = weakref.ref(layer)
        self._layer = layer

    layer = property(_get_layer, _set_layer, doc="The :class:`Layer` that this object belongs to (if it isn't a font lib). This should not be set externally.")

    def _get_glyph(self):
        if self._glyph is not None:
            return self._glyph()
        return None

    def _set_glyph(self, glyph):
        assert self._font is None
        assert self._layer is None
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._glyph = glyph

    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this object belongs to (if it isn't a font or layer lib). This should not be set externally.")

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Lib, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
from defcon import Font, Info
from defcon.test.testTools import getTestFontPath


class InfoTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.info = Info()

    def tearDown(self):
        del self.font

    def test_getParent(self):
        info = self.info
        self.assertIsNone(info.getParent())
        info = Info(self.font)
        self.assertEqual(info.getParent(), self.font)

    def test_font(self):
        info = self.info
        self.assertIsNone(info.font)
        info = Info(self.font)
        self.assertEqual(info.font, self.font)

    def test_endSelfNotificationObservation(self):
        font = self.font
        info = Info(font)
        self.assertIsNotNone(info.dispatcher)
        info.endSelfNotificationObservation()
        self.assertIsNone(info.dispatcher)
        self.assertIsNone(info.font)


if __name__ == "__main__":
    unittest.main()



import unittest
from defcon.tools.bezierMath import joinSegments


class TestBezierMath(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_joinSegments(self):
        joined = joinSegments((0, 0), (0, 138), (112, 250), (250, 250),
                              (250, 388), (500, 138), (500, 0))
        self.assertEqual(
            joined,
            ((0.0, 276.0), (500.0, 276.0), (500, 0))
        )
        # XXX in defcon/tools/bezierMath.py, the docstring was expecting
        # ((0.0, 195.16147160748713), (500.0, 471.16147160748704), (500, 0))
        # instead of ((0.0, 276.0), (500.0, 276.0), (500, 0))

    # TODO: need more tests!



import unittest
import weakref
from defcon.objects.base import BaseObject, BaseDictObject
from defcon.tools.notifications import NotificationCenter
from defcon.test.testTools import NotificationTestObserver


class TestBaseObject(BaseObject):

    def __init__(self):
        super(BaseObject, self).__init__()
        self.name = "TestBaseObject"
        self._dispatcher = None
        self.representationFactories = dict()


class BaseObjectTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_dispatcher(self):
        notificationCenter = NotificationCenter()
        obj = BaseObject()
        obj._dispatcher = weakref.ref(notificationCenter)
        self.assertEqual(obj.dispatcher, notificationCenter)

    def test_addObserver(self):
        obj = BaseObject()
        notificationObject = NotificationTestObserver()
        notificationCenter = NotificationCenter()
        obj._dispatcher = weakref.ref(notificationCenter)
        obj.addObserver(observer=notificationObject,
                        methodName="notificationCallback",
                        notification="BaseObject.Changed",
                        identifier="test")
        self.assertTrue(
            obj.dispatcher.hasObserver(observer=notificationObject,
                                       notification="BaseObject.Changed",
                                       observable=obj)
        )
        expected = [
            dict(observer=notificationObject, notification="BaseObject.Changed", observable=obj, identifier="test")
        ]
        result = obj.findObservations()
        self.assertEqual(expected, result)

    def test_removeObserver(self):
        obj = BaseObject()
        notificationObject = NotificationTestObserver()
        notificationCenter = NotificationCenter()
        obj._dispatcher = weakref.ref(notificationCenter)
        obj.dispatcher.addObserver(observer=notificationObject,
                                   methodName="notificationCallback",
                                   notification="BaseObject.Changed",
                                   observable=obj)
        obj.removeObserver(observer=notificationObject,
                           notification="BaseObject.Changed")
        self.assertFalse(
            obj.dispatcher.hasObserver(observer=notificationObject,
                                       notification="BaseObject.Changed",
                                       observable=obj)
        )

    def test_hasObserver(self):
        obj = BaseObject()
        notificationObject = NotificationTestObserver()
        notificationCenter = NotificationCenter()
        obj._dispatcher = weakref.ref(notificationCenter)
        obj.dispatcher.addObserver(observer=notificationObject,
                                   methodName="notificationCallback",
                                   notification="BaseObject.Changed",
                                   observable=obj)
        self.assertTrue(
            obj.hasObserver(observer=notificationObject,
                            notification="BaseObject.Changed")
        )

    def test_dirty_set(self):
        notificationObject = NotificationTestObserver()
        obj = TestBaseObject()
        notificationCenter = NotificationCenter()
        obj._dispatcher = weakref.ref(notificationCenter)
        obj.beginSelfNotificationObservation()
        obj.addObserver(notificationObject, "notificationCallback",
                        "BaseObject.Changed")
        obj.dirty = True
        self.assertEqual(len(notificationObject.stack), 1)
        self.assertEqual(notificationObject.stack[-1],
                         ("BaseObject.Changed", "TestBaseObject"))
        self.assertTrue(obj.dirty)
        obj.dirty = False
        self.assertEqual(notificationObject.stack[-1],
                         ("BaseObject.Changed", "TestBaseObject"))

    def test_dirty_get(self):
        obj = TestBaseObject()
        obj.dirty = True
        self.assertTrue(obj.dirty)
        obj.dirty = False
        self.assertFalse(obj.dirty)


def _representationTestFactory(obj, **kwargs):
    return repr(tuple(sorted(kwargs.items())))


def sortRepresentationKeys(reprensentationKeys):
    return [
        (k, sorted((kk, vv) for kk, vv in v.items()))
        for k, v in sorted(reprensentationKeys,
                           key=lambda t: (t[0], sorted(t[1].keys())))]


class RepresentationsTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.obj = BaseObject()

    def tearDown(self):
        del self.obj

    def test_object_representations(self):
        notificationCenter = NotificationCenter()
        self.obj._dispatcher = weakref.ref(notificationCenter)
        self.obj.representationFactories = dict(
            test=dict(
                factory=_representationTestFactory,
                destructiveNotifications=["BaseObject.Changed"]))
        self.assertEqual(self.obj.getRepresentation("test"), "()")
        self.assertEqual(
            self.obj.representationKeys(),
            [('test', {})]
        )
        self.assertEqual(
            self.obj.getRepresentation("test", attr1="foo",
                                       attr2="bar", attr3=1),
            "(('attr1', 'foo'), ('attr2', 'bar'), ('attr3', 1))"
        )
        self.assertEqual(sortRepresentationKeys(self.obj.representationKeys()),
                         sorted([('test', []),
                                 ('test', [('attr1', 'foo'), ('attr2', 'bar'),
                                           ('attr3', 1)])]))
        self.assertTrue(self.obj.hasCachedRepresentation("test"))
        self.assertTrue(self.obj.hasCachedRepresentation("test", attr1="foo",
                                                         attr2="bar", attr3=1))
        self.assertFalse(self.obj.hasCachedRepresentation("test",
                                                          attr1="not foo",
                                                          attr2="bar",
                                                          attr3=1))
        self.obj.destroyAllRepresentations()
        self.assertEqual(self.obj.representationKeys(),
                         [])
        self.obj.representationFactories["foo"] = dict(
            factory=_representationTestFactory,
            destructiveNotifications=["BaseObject.Changed"]
        )
        self.assertEqual(self.obj.getRepresentation("test"),
                         '()')
        self.assertEqual(self.obj.getRepresentation("test", attr1="foo",
                                                    attr2="bar", attr3=1),
                         "(('attr1', 'foo'), ('attr2', 'bar'), ('attr3', 1))")
        self.assertEqual(
            self.obj.getRepresentation("test", attr21="foo",
                                       attr22="bar", attr23=1),
            "(('attr21', 'foo'), ('attr22', 'bar'), ('attr23', 1))")
        self.assertEqual(self.obj.getRepresentation("foo"),
                         "()")
        self.obj.destroyRepresentation("test", attr21="foo",
                                       attr22="bar", attr23=1)
        self.assertEqual(sortRepresentationKeys(self.obj.representationKeys()),
                         [('foo', []),
                          ('test', []),
                          ('test', [('attr1', 'foo'), ('attr2', 'bar'),
                                    ('attr3', 1)])])
        self.obj.destroyRepresentation("test")
        self.assertEqual(self.obj.representationKeys(),
                         [('foo', {})])

    def test_object_representations_no_dispatcher(self):
        self.obj.representationFactories = dict(
            test=dict(
                factory=_representationTestFactory,
                destructiveNotifications=["BaseObject.Changed"]))
        self.assertEqual(self.obj.getRepresentation("test"), "()")
        self.assertEqual(
            [],
            []
        )

class TestBaseDictObject(BaseDictObject):

    def __init__(self):
        super(BaseDictObject, self).__init__()
        self.name = "TestBaseDictObject"
        self._dispatcher = None
        self.representationFactories = dict()


class BaseDictObjectTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.obj = BaseDictObject()

    def tearDown(self):
        pass

    def test_object_contains(self):
        self.obj["A"] = 1
        self.assertTrue("A" in self.obj)
        self.assertFalse("B" in self.obj)

    def test_len(self):
        self.assertEqual(len(self.obj), 0)
        self.obj["A"] = 1
        self.assertEqual(len(self.obj), 1)

    def test_getItem(self):
        self.obj["A"] = 1
        self.assertEqual(self.obj["A"], 1)

    def test_setItem(self):
        self.obj["A"] = 1
        self.assertEqual(self.obj["A"], 1)
        self.assertTrue(self.obj.dirty)

    def test_delItem(self):
        self.obj["A"] = 1
        self.obj.dirty = False
        del self.obj["A"]
        self.assertFalse("A" in self.obj)
        self.assertTrue(self.obj.dirty)

    def test_get(self):
        self.obj["A"] = 1
        self.assertEqual(self.obj.get("A"), 1)
        self.assertEqual(self.obj.get("B"), None)

    def test_clear(self):
        self.obj["A"] = 1
        self.obj.dirty = False
        self.obj.clear()
        self.assertEqual(len(self.obj), 0)
        self.assertTrue(self.obj.dirty, True)

    def test_update(self):
        self.obj["A"] = 1
        self.obj.dirty = False
        self.obj.update(dict(A=2, B=3))
        self.assertEqual(len(self.obj), 2)
        self.assertEqual(self.obj["A"], 2)
        self.assertEqual(self.obj["B"], 3)
        self.assertEqual(self.obj.dirty, True)

    def test_keys(self):
        self.obj["A"] = 1
        self.assertEqual(list(self.obj.keys()), ["A"])

    def test_values(self):
        self.obj["A"] = 1
        self.assertEqual(list(self.obj.values()), [1])

    def test_items(self):
        self.obj["A"] = 1
        self.assertEqual(list(self.obj.items()), [("A", 1)])

    def test_dirty(self):
        self.assertFalse(self.obj.dirty)
        notdirty = not self.obj.dirty
        self.obj.dirty = notdirty
        self.assertEqual(self.obj.dirty, notdirty)
        self.obj.dirty = not notdirty
        self.assertNotEqual(self.obj.dirty, notdirty)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
# this file was generated by infoObjectGenerator.py.
# this file should not be edited by hand.

import weakref
from warnings import warn
from fontTools import ufoLib
from defcon.objects.base import BaseObject
from copy import copy
from functools import partial

def _guidelineDeprecation(message=None):
    msg = "Font level guidelines are now handled in the Font object."
    if message is not None:
        msg += " " + message
    warn(msg, DeprecationWarning, stacklevel=2)

def init_property(cls, name, setup):
    setterName = '_set_{0}'.format(name)
    getterName = '_get_{0}'.format(name)
    privateName = '_{0}'.format(name)

    doc, default = setup

    def getter(self):
        return getattr(self, privateName)
    getter.__name__ = getterName

    def setter(self, value):
        oldValue = getattr(self, privateName)
        if oldValue == value:
            return
        if value is None:
            value = copy(default)
        else:
            valid = ufoLib.validateFontInfoVersion3ValueForAttribute(name, value)
            if not valid:
                raise ValueError("Invalid value ({0}) for attribute {1}.".format(repr(value), name))
        setattr(self, privateName, value)
        self.postNotification("Info.ValueChanged", data=dict(attribute=name, oldValue=oldValue, newValue=value))
        self.dirty = True
    setter.__name__ = setterName

    prop = property(getter, setter, doc)

    setattr(cls, setterName, setter)
    setattr(cls, getterName, getter)
    setattr(cls, name, prop)

def init_properties(cls):
    for name in cls._properties:
        init_property(cls, name, cls._properties[name])
    return cls

@init_properties
class Info(BaseObject):
    """
    This object represents info values.

    **This object posts the following notifications:**

    - Info.Changed
    - Info.BeginUndo
    - Info.EndUndo
    - Info.BeginRedo
    - Info.EndRedo
    - Info.ValueChanged

    **Note:** The documentation strings here were automatically generated
    from the `UFO specification <http://unifiedfontobject.org/filestructure/fontinfo.html>`_.
    """

    changeNotificationName = "Info.Changed"
    beginUndoNotificationName = "Info.BeginUndo"
    endUndoNotificationName = "Info.EndUndo"
    beginRedoNotificationName = "Info.BeginRedo"
    endRedoNotificationName = "Info.EndRedo"
    representationFactories = {}

    def __init__(self, font=None, guidelineClass=None):
        if font is not None:
            font = weakref.ref(font)
        self._font = font
        super(Info, self).__init__()
        self.beginSelfNotificationObservation()
        if guidelineClass is not None:
            _guidelineDeprecation()
        # init private property attributes
        for name in self._properties:
            _, default = self._properties[name]
            setattr(self, '_'+name, copy(default))
        self._dirty = False

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # ----------
    # Properties
    # ----------
    _properties = {
        "ascender": ("Ascender value. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "capHeight": ("Cap height value. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "copyright": ("Copyright statement. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "descender": ("Descender value. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "familyName": ("Family name. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "italicAngle": ("Italic angle. This must be an angle in counter-clockwise degrees from the vertical. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "macintoshFONDFamilyID": ("Family ID number. Corresponds to the ffFamID in the FOND resource. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "macintoshFONDName": ("Font name for the FOND resource. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "note": ("Arbitrary note about the font. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeGaspRangeRecords": ("A list of gasp Range Records. These must be sorted in ascending order based on the <code>rangeMaxPPEM</code> value of the record. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHeadCreated": ("Creation date. Expressed as a string of the format \"YYYY/MM/DD HH:MM:SS\". \"YYYY/MM/DD\" is year/month/day. The month must be in the range 1-12 and the day must be in the range 1-end of month. \"HH:MM:SS\" is hour:minute:second. The hour must be in the range 0:23. The minute and second must each be in the range 0-59. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHeadFlags": ("A list of bit numbers indicating the flags. The bit numbers are listed in the OpenType head specification. Corresponds to the OpenType head table <code>flags</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHeadLowestRecPPEM": ("Smallest readable size in pixels. Corresponds to the OpenType head table <code>lowestRecPPEM</code> field. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaAscender": ("Ascender value. Corresponds to the OpenType hhea table <code>Ascender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaCaretOffset": ("Caret offset value. Corresponds to the OpenType hhea table <code>caretOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaCaretSlopeRise": ("Caret slope rise value. Corresponds to the OpenType hhea table <code>caretSlopeRise</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaCaretSlopeRun": ("Caret slope run value. Corresponds to the OpenType hhea table <code>caretSlopeRun</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaDescender": ("Descender value. Corresponds to the OpenType hhea table <code>Descender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeHheaLineGap": ("Line gap value. Corresponds to the OpenType hhea table <code>LineGap</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameCompatibleFullName": ("Compatible full name. Corresponds to the OpenType name table name ID 18. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameDescription": ("Description of the font. Corresponds to the OpenType name table name ID 10. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameDesigner": ("Designer name. Corresponds to the OpenType name table name ID 9. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameDesignerURL": ("URL for the designer. Corresponds to the OpenType name table name ID 12. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameLicense": ("License text. Corresponds to the OpenType name table name ID 13. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameLicenseURL": ("URL for the license. Corresponds to the OpenType name table name ID 14. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameManufacturer": ("Manufacturer name. Corresponds to the OpenType name table name ID 8. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameManufacturerURL": ("Manufacturer URL. Corresponds to the OpenType name table name ID 11. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNamePreferredFamilyName": ("Preferred family name. Corresponds to the OpenType name table name ID 16. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNamePreferredSubfamilyName": ("Preferred subfamily name. Corresponds to the OpenType name table name ID 17. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameRecords": ("A list of name records. This name record storage area is intended for records that require platform, encoding and or language localization. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameSampleText": ("Sample text. Corresponds to the OpenType name table name ID 19. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameUniqueID": ("Unique ID string. Corresponds to the OpenType name table name ID 3. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameVersion": ("Version string. Corresponds to the OpenType name table name ID 5. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameWWSFamilyName": ("WWS family name. Corresponds to the OpenType name table name ID 21. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeNameWWSSubfamilyName": ("WWS Subfamily name. Corresponds to the OpenType name table name ID 22. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2CodePageRanges": ("A list of bit numbers that are supported code page ranges in the font. The bit numbers are listed in the OpenType OS/2 specification. Corresponds to the OpenType OS/2 table <code>ulCodePageRange1</code> and <code>ulCodePageRange2</code> fields. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2FamilyClass": ("Two integers representing the IBM font class and font subclass of the font. The first number, representing the class ID, must be in the range 0-14. The second number, representing the subclass, must be in the range 0-15. The numbers are listed in the OpenType OS/2 specification. Corresponds to the OpenType OS/2 table <code>sFamilyClass</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2Panose": ("The list must contain 10 non-negative integers that represent the setting for each category in the Panose specification. The integers correspond with the option numbers in each of the Panose categories. This corresponds to the OpenType OS/2 table <code>Panose</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2Selection": ("A list of bit numbers indicating the bits that should be set in fsSelection. The bit numbers are listed in the OpenType OS/2 specification. Corresponds to the OpenType OS/2 table <code>selection</code> field. Note: Bits 0 (italic), 5 (bold) and 6 (regular) must not be set here. These bits should be taken from the generic styleMapStyle attribute. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2StrikeoutPosition": ("Strikeout position. Corresponds to the OpenType OS/2 table <code>yStrikeoutPosition</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2StrikeoutSize": ("Strikeout size. Corresponds to the OpenType OS/2 table <code>yStrikeoutSize</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SubscriptXOffset": ("Subscript x offset. Corresponds to the OpenType OS/2 table <code>ySubscriptXOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SubscriptXSize": ("Subscript horizontal font size. Corresponds to the OpenType OS/2 table <code>ySubscriptXSize</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SubscriptYOffset": ("Subscript y offset. Corresponds to the OpenType OS/2 table <code>ySubscriptYOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SubscriptYSize": ("Subscript vertical font size. Corresponds to the OpenType OS/2 table <code>ySubscriptYSize</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SuperscriptXOffset": ("Superscript x offset. Corresponds to the OpenType OS/2 table <code>ySuperscriptXOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SuperscriptXSize": ("Superscript horizontal font size. Corresponds to the OpenType OS/2 table <code>ySuperscriptXSize</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SuperscriptYOffset": ("Superscript y offset. Corresponds to the OpenType OS/2 table <code>ySuperscriptYOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2SuperscriptYSize": ("Superscript vertical font size. Corresponds to the OpenType OS/2 table <code>ySuperscriptYSize</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2Type": ("A list of bit numbers indicating the embedding type. The bit numbers are listed in the OpenType OS/2 specification. Corresponds to the OpenType OS/2 table <code>fsType</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2TypoAscender": ("Ascender value. Corresponds to the OpenType OS/2 table <code>sTypoAscender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2TypoDescender": ("Descender value. Corresponds to the OpenType OS/2 table <code>sTypoDescender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2TypoLineGap": ("Line gap value. Corresponds to the OpenType OS/2 table <code>sTypoLineGap</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2UnicodeRanges": ("A list of bit numbers that are supported Unicode ranges in the font. The bit numbers are listed in the OpenType OS/2 specification. Corresponds to the OpenType OS/2 table <code>ulUnicodeRange1</code>, <code>ulUnicodeRange2</code>, <code>ulUnicodeRange3</code> and <code>ulUnicodeRange4</code> fields. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2VendorID": ("Four character identifier for the creator of the font. Corresponds to the OpenType OS/2 table <code>achVendID</code> field. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2WeightClass": ("Weight class value. Must be a non-negative integer. Corresponds to the OpenType OS/2 table <code>usWeightClass</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2WidthClass": ("Width class value. Must be in the range 1-9. Corresponds to the OpenType OS/2 table <code>usWidthClass</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2WinAscent": ("Ascender value. Corresponds to the OpenType OS/2 table <code>usWinAscent</code> field. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeOS2WinDescent": ("Descender value. Corresponds to the OpenType OS/2 table <code>usWinDescent</code> field. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaCaretOffset": ("Caret offset value. Corresponds to the OpenType vhea table <code>caretOffset</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaCaretSlopeRise": ("Caret slope rise value. Corresponds to the OpenType vhea table <code>caretSlopeRise</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaCaretSlopeRun": ("Caret slope run value. Corresponds to the OpenType vhea table <code>caretSlopeRun</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaVertTypoAscender": ("Ascender value. Corresponds to the OpenType vhea table <code>vertTypoAscender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaVertTypoDescender": ("Descender value. Corresponds to the OpenType vhea table <code>vertTypoDescender</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "openTypeVheaVertTypoLineGap": ("Line gap value. Corresponds to the OpenType vhea table <code>vertTypoLineGap</code> field. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "postscriptBlueFuzz": ("BlueFuzz value. This corresponds to the Type 1/CFF <code>BlueFuzz</code> field. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptBlueScale": ("BlueScale value. This corresponds to the Type 1/CFF <code>BlueScale</code> field. This should be a float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptBlueShift": ("BlueShift value. This corresponds to the Type 1/CFF <code>BlueShift</code> field. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptBlueValues": ("A list of up to 14 integers or floats specifying the values that should be in the Type 1/CFF BlueValues field. This list must contain an even number of integers following the rules defined in the Type 1/CFF specification. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptDefaultCharacter": ("The name of the glyph that should be used as the default character in PFM files. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "postscriptDefaultWidthX": ("Default width for glyphs. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptFamilyBlues": ("A list of up to 14 integers or floats specifying the values that should be in the Type 1/CFF FamilyBlues field. This list must contain an even number of integers following the rules defined in the Type 1/CFF specification. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptFamilyOtherBlues": ("A list of up to 10 integers or floats specifying the values that should be in the Type 1/CFF FamilyOtherBlues field. This list must contain an even number of integers following the rules defined in the Type 1/CFF specification. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptFontName": ("Name to be used for the <code>FontName</code> field in Type 1/CFF table. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "postscriptForceBold": ("Indicates how the Type 1/CFF <code>ForceBold</code> field should be set. This should be a boolean. Setting this will post an *Info.Changed* notification.", None),
        "postscriptFullName": ("Name to be used for the <code>FullName</code> field in Type 1/CFF table. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "postscriptIsFixedPitch": ("Indicates if the font is monospaced. An authoring tool could calculate this automatically, but the designer may wish to override this setting. This corresponds to the Type 1/CFF <code>isFixedPitched</code> field This should be a boolean. Setting this will post an *Info.Changed* notification.", None),
        "postscriptNominalWidthX": ("Nominal width for glyphs. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptOtherBlues": ("A list of up to 10 integers or floats specifying the values that should be in the Type 1/CFF OtherBlues field. This list must contain an even number of integers following the rules defined in the Type 1/CFF specification. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptSlantAngle": ("Artificial slant angle. This must be an angle in counter-clockwise degrees from the vertical. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptStemSnapH": ("List of horizontal stems sorted in the order specified in the Type 1/CFF specification. Up to 12 integers or floats are possible. This corresponds to the Type 1/CFF <code>StemSnapH</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptStemSnapV": ("List of vertical stems sorted in the order specified in the Type 1/CFF specification. Up to 12 integers or floats are possible. This corresponds to the Type 1/CFF <code>StemSnapV</code> field. This should be a list. Setting this will post an *Info.Changed* notification.", []),
        "postscriptUnderlinePosition": ("Underline position value. Corresponds to the Type 1/CFF/post table <code>UnderlinePosition</code> field. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptUnderlineThickness": ("Underline thickness value. Corresponds to the Type 1/CFF/post table <code>UnderlineThickness</code> field. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "postscriptUniqueID": ("A unique ID number as defined in the Type 1/CFF specification. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "postscriptWeightName": ("A string indicating the overall weight of the font. This corresponds to the Type 1/CFF Weight field. It should be in sync with the <code>openTypeOS2WeightClass</code> value. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "postscriptWindowsCharacterSet": ("The Windows character set. The values are defined below. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "styleMapFamilyName": ("Family name used for bold, italic and bold italic style mapping. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "styleMapStyleName": ("Style map style. The possible values are regular, italic, bold and bold italic. These are case sensitive. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "styleName": ("Style name. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "trademark": ("Trademark statement. This should be a string. Setting this will post an *Info.Changed* notification.", None),
        "unitsPerEm": ("Units per em. This should be a non-negative integer or float. Setting this will post an *Info.Changed* notification.", None),
        "versionMajor": ("Major version. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
        "versionMinor": ("Minor version. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "woffMajorVersion": ("Major version of the font. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataCopyright": ("Font copyright. Corresponds to the WOFF <code>copyright</code> element. The dictionary must follow the WOFF Metadata Copyright Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataCredits": ("Font credits. Corresponds to the WOFF <code>credits</code> element. The dictionary must follow the WOFF Metadata Credits Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataDescription": ("Font description. Corresponds to the WOFF <code>description</code> element. The dictionary must follow the WOFF Metadata Description Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataExtensions": ("List of metadata extension records. The dictionaries must follow the WOFF Metadata Extension Record structure. There must be at least one extension record in the list. This should be a list. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataLicense": ("Font description. Corresponds to the WOFF <code>license</code> element. The dictionary must follow the WOFF Metadata License Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataLicensee": ("Font licensee. Corresponds to the WOFF <code>licensee</code> element. The dictionary must follow the WOFF Metadata Licensee Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataTrademark": ("Font trademark. Corresponds to the WOFF <code>trademark</code> element. The dictionary must follow the WOFF Metadata Trademark Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataUniqueID": ("Identification string. Corresponds to the WOFF <code>uniqueid</code>. The dictionary must follow the WOFF Metadata Unique ID Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMetadataVendor": ("Font vendor. Corresponds to the WOFF <code>vendor</code> element. The dictionary must follow the the WOFF Metadata Vendor Record structure. This should be a dictionary. Setting this will post an *Info.Changed* notification.", None),
        "woffMinorVersion": ("Minor version of the font. This should be a non-negative integer. Setting this will post an *Info.Changed* notification.", None),
        "xHeight": ("x-height value. This should be a integer or float. Setting this will post an *Info.Changed* notification.", None),
        "year": ("The year the font was created. This attribute is deprecated as of version 2. It's presence should not be relied upon by authoring tools. However, it may occur in a font's info so authoring tools should preserve it if present. This should be a integer. Setting this will post an *Info.Changed* notification.", None),
    }

    # ----------
    # Guidelines
    # ----------

    def _get_guidelines(self):
        font = self.font
        if font is None:
            return []
        return font.guidelines

    def _set_guidelines(self, value):
        font = self.font
        if font is None:
            return []
        font.guidelines = value

    guidelines = property(_get_guidelines, _set_guidelines, doc="This is a compatibility attribute for ufoLib. It maps to :py:attr:`Font.guidelines`.")

    def appendGuideline(self, *args, **kwargs):
        _guidelineDeprecation("This maps to Font.appendGuideline.")
        font = self.font
        if font is None:
            raise NotImplementedError
        return font.appendGuideline(*args, **kwargs)

    def insertGuideline(self, *args, **kwargs):
        _guidelineDeprecation("This maps to Font.insertGuideline.")
        font = self.font
        if font is None:
            raise NotImplementedError
        return font.insertGuideline(*args, **kwargs)

    def removeGuideline(self, *args, **kwargs):
        _guidelineDeprecation("This maps to Font.removeGuideline.")
        font = self.font
        if font is None:
            raise NotImplementedError
        return font.removeGuideline(*args, **kwargs)

    def guidelineIndex(self, *args, **kwargs):
        _guidelineDeprecation("This maps to Font.guidelineIndex.")
        font = self.font
        if font is None:
            raise NotImplementedError
        return font.guidelineIndex(*args, **kwargs)

    def clearGuidelines(self, *args, **kwargs):
        _guidelineDeprecation("This maps to Font.clearGuidelines.")
        font = self.font
        if font is None:
            raise NotImplementedError
        font.clearGuidelines(*args, **kwargs)

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        if self.dispatcher is None:
            return
        super(Info, self).endSelfNotificationObservation()
        self._font = None

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        simple_get = partial(getattr, self)

        getters = []
        for name in self._properties:
            if getattr(self, '_' + name) is None:
                continue
            getters.append((name, simple_get))

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        simple_set = partial(setattr, self)

        setters = [(name, simple_set) for name in self._properties]

        for name, setter in setters:
            if name not in data:
                continue
            setter(name, data[name])



import unittest
from defcon import Image, Font
from defcon.test.testTools import (
    getTestFontPath, makeTestFontCopy, tearDownTestFontCopy)


class ImageTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_transformation(self):
        i = Image()
        i.transformation = (1, 2, 3, 4, 5, 6)
        self.assertEqual(i.transformation, (1, 2, 3, 4, 5, 6))
        self.assertTrue(i.dirty)

    def test_color(self):
        i = Image()
        i.color = "1, 1, 1, 1"
        self.assertEqual(i.color, "1,1,1,1")
        self.assertTrue(i.dirty)

    def test_dirty(self):
        i = Image()
        self.assertFalse(i.dirty)

    def test_fileName(self):
        i = Image()
        i.fileName = "foo"
        self.assertEqual(i.fileName, "foo")
        self.assertTrue(i.dirty)

    def test_instance(self):
        i = Image(imageDict=dict(fileName="foo.png", xScale="1", xyScale="2",
                                 yxScale="3", yScale="4", xOffset="5",
                                 yOffset="6", color="0,0,0,0"))
        self.assertEqual(
            (i.fileName, i.transformation, i.color),
            ('foo.png', ('1', '2', '3', '4', '5', '6'), '0,0,0,0'))

    def test_clear(self):
        i = Image()
        i.fileName = "foo"
        self.assertEqual(i.fileName, "foo")
        i.clear()
        self.assertIsNone(i.fileName)

    def test_read(self):
        font = Font(getTestFontPath())
        glyph = font.layers["Layer 1"]["A"]
        image = glyph.image
        self.assertEqual(image.fileName, 'image 1.png')
        self.assertEqual(image.color, '0.1,0.2,0.3,0.4')
        self.assertEqual(image.transformation, (0.5, 0, 0, 0.5, 0, 0))

    def test_write(self):
        path = makeTestFontCopy()
        font = Font(path)
        glyph = font.layers[None]["A"]
        glyph.image = glyph.instantiateImage()
        glyph.image.color = "1,1,1,1"
        glyph.image.fileName = "foo.png"
        glyph.image.transformation = (1, 2, 3, 4, 5, 6)
        font.save()
        font = Font(path)
        glyph = font.layers[None]["A"]
        self.assertEqual(sorted(glyph.image.items()),
                         [('color', '1,1,1,1'), ('fileName', 'foo.png'),
                          ('xOffset', 5), ('xScale', 1), ('xyScale', 2),
                          ('yOffset', 6), ('yScale', 4), ('yxScale', 3)])
        tearDownTestFontCopy()


if __name__ == "__main__":
    unittest.main()



import unittest
from defcon.objects.font import Font
from defcon.test.testTools import getTestFontPath


class UnicodeDataTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_removeGlyphData(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("XXX")
        font.unicodeData.addGlyphData("XXX", [65])
        font.unicodeData.removeGlyphData("A", [65])
        self.assertEqual(font.unicodeData[65], ['XXX'])

    def test_addGlyphData(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("XXX")
        font.unicodeData.addGlyphData("XXX", [1000])
        self.assertEqual(font.unicodeData[1000], ['XXX'])
        font.unicodeData.addGlyphData("XXX", [65])
        self.assertEqual(font.unicodeData[65], ['A', 'XXX'])

    def test_delitem(self):
        path = getTestFontPath()
        font = Font(path)
        del font.unicodeData[65]
        self.assertNotIn(65, font.unicodeData)
        font.unicodeData.glyphNameForUnicode(65)

        self.assertNotIn(0xBEAF, font.unicodeData)
        del font.unicodeData[0xBEAF]
        self.assertNotIn(0xBEAF, font.unicodeData)

    def test_setitem(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("XXX")
        font.unicodeData[1000] = ["XXX"]
        self.assertEqual(font.unicodeData[1000], ['XXX'])
        font.unicodeData[65] = ["YYY"]
        self.assertEqual(font.unicodeData[65], ['A', 'YYY'])

    def test_clear(self):
        path = getTestFontPath()
        font = Font(path)
        font.unicodeData.clear()
        self.assertEqual(list(font.unicodeData.keys()), [])

    def test_update(self):
        path = getTestFontPath()
        font1 = Font(path)
        font2 = Font()
        font2.unicodeData.update(font1.unicodeData)
        self.assertEqual(font1.unicodeData, font2.unicodeData)

    def test_unicodeForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.unicodeData.unicodeForGlyphName("A"), 65)

    def test_glyphNameForUnicode(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.unicodeData.glyphNameForUnicode(65), 'A')

    def test_pseudoUnicodeForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.unicodeData.pseudoUnicodeForGlyphName("A"), 65)
        font.newGlyph("A.foo")
        self.assertEqual(font.unicodeData.pseudoUnicodeForGlyphName("A.foo"),
                         65)
        font.newGlyph("B_A")
        self.assertEqual(font.unicodeData.pseudoUnicodeForGlyphName("B_A"), 66)

    def test_forcedUnicodeForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.unicodeData.forcedUnicodeForGlyphName("A"), 65)
        font.newGlyph("B_A")
        self.assertEqual(font.unicodeData.forcedUnicodeForGlyphName("B_A"),
                         0xE000)
        font.newGlyph("B_B")
        self.assertEqual(font.unicodeData.forcedUnicodeForGlyphName("B_B"),
                         0xE001)

    def test_glyphNameForForcedUnicode(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.unicodeData.glyphNameForForcedUnicode(65), "A")
        font.newGlyph("B_A")
        self.assertIsNone(font.unicodeData.glyphNameForForcedUnicode(0xE000))
        font.unicodeData.forcedUnicodeForGlyphName("B_A")
        self.assertEqual(font.unicodeData.glyphNameForForcedUnicode(0xE000),
                         "B_A")
        font.newGlyph("B_B")
        font.unicodeData.forcedUnicodeForGlyphName("B_B")
        self.assertEqual(font.unicodeData.glyphNameForForcedUnicode(0xE001),
                         "B_B")

    def test_scriptForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("A.alt")
        self.assertEqual(font.unicodeData.scriptForGlyphName("A"), 'Latin')
        self.assertEqual(font.unicodeData.scriptForGlyphName("A.alt"), 'Latin')
        self.assertEqual(font.unicodeData.scriptForGlyphName("A.alt", False),
                         'Unknown')
        font.newGlyph("Alpha")
        font["Alpha"].unicode = 0x0391
        self.assertEqual(font.unicodeData.scriptForGlyphName("Alpha"), 'Greek')

    def test_blockForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("A.alt")
        self.assertEqual(font.unicodeData.blockForGlyphName("A"),
                         'Basic Latin')
        self.assertEqual(font.unicodeData.blockForGlyphName("A.alt"),
                         'Basic Latin')
        self.assertEqual(font.unicodeData.blockForGlyphName("A.alt", False),
                         'No_Block')
        font.newGlyph("schwa")
        font["schwa"].unicode = 0x0259
        self.assertEqual(font.unicodeData.blockForGlyphName("schwa"),
                         'IPA Extensions')

    def test_categoryForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("A.alt")
        self.assertEqual(font.unicodeData.categoryForGlyphName("A"), 'Lu')
        self.assertEqual(font.unicodeData.categoryForGlyphName("A.alt"), 'Lu')
        self.assertEqual(font.unicodeData.categoryForGlyphName("A.alt", False),
                         'Cn')

    def test_decompositionBaseForGlyphName(self):
        path = getTestFontPath()
        font = Font(path)
        font.newGlyph("Aacute")
        font["Aacute"].unicode = int("00C1", 16)
        self.assertEqual(
            font.unicodeData.decompositionBaseForGlyphName("Aacute", True),
            'A')
        font.newGlyph("Aringacute")
        font["Aringacute"].unicode = int("01FA", 16)
        self.assertEqual(
            font.unicodeData.decompositionBaseForGlyphName("Aringacute", True),
            'A')
        font.newGlyph("Aacute.alt")
        self.assertEqual(
            font.unicodeData.decompositionBaseForGlyphName("Aacute.alt", True),
            'A')
        font.newGlyph("A.alt")
        self.assertEqual(
            font.unicodeData.decompositionBaseForGlyphName("Aacute.alt", True),
            'A.alt')

    def test_closeRelativeForGlyphName(self):
        font = Font()
        font.newGlyph("parenleft")
        font["parenleft"].unicode = int("0028", 16)
        font.newGlyph("parenright")
        font["parenright"].unicode = int("0029", 16)
        font.newGlyph("parenleft.alt")
        font.newGlyph("parenright.alt")
        self.assertEqual(
            font.unicodeData.closeRelativeForGlyphName("parenleft", True),
            'parenright')
        self.assertEqual(
            font.unicodeData.closeRelativeForGlyphName("parenleft.alt", True),
            'parenright.alt')
        del font["parenright.alt"]
        self.assertEqual(
            font.unicodeData.closeRelativeForGlyphName("parenleft.alt", True),
            'parenright')

    def test_openRelativeForGlyphName(self):
        font = Font()
        font.newGlyph("parenleft")
        font["parenleft"].unicode = int("0028", 16)
        font.newGlyph("parenright")
        font["parenright"].unicode = int("0029", 16)
        font.newGlyph("parenleft.alt")
        font.newGlyph("parenright.alt")
        self.assertEqual(
            font.unicodeData.openRelativeForGlyphName("parenright", True),
            'parenleft')
        self.assertEqual(
            font.unicodeData.openRelativeForGlyphName("parenright.alt", True),
            'parenleft.alt')
        del font["parenleft.alt"]
        self.assertEqual(
            font.unicodeData.openRelativeForGlyphName("parenright.alt", True),
            'parenleft')


class SortGlyphNamesTest(unittest.TestCase):

    def setUp(self):
        self.font = font = Font()
        font.newGlyph("a")
        font["a"].unicode = 0x0061
        font.newGlyph("b")
        font["b"].unicode = 0x0062
        font.newGlyph("c")
        font["c"].unicode = 0x0063
        font.newGlyph("alpha")
        font["alpha"].unicode = 0x03B1
        font.newGlyph("aacute")
        font["aacute"].unicode = 0x00E1
        font.newGlyph("comma")
        font["comma"].unicode = 0x002C
        font.newGlyph("schwa")
        font["schwa"].unicode = 0x0259
        font.newGlyph("undefined")

    def tearDown(self):
        del self.font

    def test_sortGlyphNames_unicode(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(["a", "c", "b"]),
            ["a", "b", "c"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a.suf1", "a.suf2", "a", "b"]),
            ["a", "b", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a.suf1", "a.suf2", "a", "b"],
                sortDescriptors=[dict(type="unicode", allowPseudoUnicode=True)]
            ),
            ["a", "a.suf1", "a.suf2", "b"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a", "b", "c"],
                sortDescriptors=[dict(type="unicode", ascending=False)]
            ),
            ["c", "b", "a"]
        )

    def test_sortGlyphNames_alphabetical(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "c", "a.suf1", "a.suf2", "a"],
                sortDescriptors=[dict(type="alphabetical")]
            ),
            ["a", "a.suf1", "a.suf2", "b", "c"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a.suf1", "a.suf2", "a", "b", "c"],
                sortDescriptors=[dict(type="alphabetical", ascending=False)]
            ),
            ["c", "b", "a.suf2", "a.suf1", "a"]
        )

    def test_sortGlyphNames_multiple_sortDescriptors(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a.suf1", "a.suf2", "a", "b", "c"],
                sortDescriptors=[
                    dict(type="alphabetical", ascending=False),
                    dict(type="suffix")
                ]
            ),
            ["c", "b", "a", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a.suf1", "a.suf2", "a", "b", "c"],
                sortDescriptors=[
                    dict(type="alphabetical"),
                    dict(type="suffix", ascending=False)
                ]
            ),
            ["a", "b", "c", "a.suf1", "a.suf2"]
            # XXX: ["a.suf2", "a.suf1", "a", "b", "c"]
        )

    def test_sortGlyphNames_script(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "alpha", "a"],
                sortDescriptors=[dict(type="script")]
            ),
            ["b", "a", "alpha"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "alpha", "a"],
                sortDescriptors=[dict(type="script", ascending=False)]
            ),
            ["alpha", "b", "a"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "alpha", "a", "a.suf1", "a.suf2"],
                sortDescriptors=[dict(type="script")]
            ),
            ["b", "a", "alpha", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "alpha", "a", "a.suf1", "a.suf2"],
                sortDescriptors=[dict(type="script", allowPseudoUnicode=True)]
            ),
            ["b", "a", "a.suf1", "a.suf2", "alpha"]
        )

    def test_sortGlyphNames_category(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "comma", "a"],
                sortDescriptors=[dict(type="category")]
            ),
            ["b", "a", "comma"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "comma", "a"],
                sortDescriptors=[dict(type="category", ascending=False)]
            ),
            ["comma", "b", "a"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf1", "a.suf2", "comma", "a"],
                sortDescriptors=[
                    dict(type="category")
                ]
            ),
            ["b", "a", "comma", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf1", "a.suf2", "comma", "a"],
                sortDescriptors=[
                    dict(type="category", allowPseudoUnicode=True)
                ]
            ),
            ["b", "a.suf1", "a.suf2", "a", "comma"]
        )

    def test_sortGlyphNames_block(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "schwa", "a", "undefined"],
                sortDescriptors=[dict(type="block")]
            ),
            ["b", "a", "schwa"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "schwa", "a", "undefined"],
                sortDescriptors=[dict(type="block", ascending=False)]
            ),
            ["schwa", "b", "a"]
        )

    def test_sortGlyphNames_suffix(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf2", "a", "a.suf1", "c"],
                sortDescriptors=[dict(type="suffix")]
            ),
            ["b", "a", "c", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf2", "a", "a.suf1", "c"],
                sortDescriptors=[dict(type="suffix", ascending=False)]
            ),
            ["b", "a", "c", "a.suf1", "a.suf2"]
            # XXX: ["a.suf2", "a.suf1", "b", "a", "c"]?
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf2", "a", "a.suf1", "c"],
                sortDescriptors=[dict(type="suffix", allowPseudoUnicode=True)]
            ),
            ["b", "a", "c", "a.suf1", "a.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a.suf2", "a", "a.suf1", "c"],
                sortDescriptors=[
                    dict(type="suffix", ascending=False,
                         allowPseudoUnicode=True)]
            ),
            ["b", "a", "c", "a.suf1", "a.suf2"]
            # XXX: ["a.suf2", "a.suf1", "b", "a", "c"]?
        )

    def test_sortGlyphNames_decompositionBase(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a", "a.alt", "aacute", "aacute.alt"],
                sortDescriptors=[dict(type="decompositionBase")]
            ),
            ["a", "aacute", "a.alt", "aacute.alt"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a", "a.alt", "aacute", "aacute.alt"],
                sortDescriptors=[
                    dict(type="decompositionBase", ascending=False)]
            ),
            ["aacute.alt", "a.alt", "a", "aacute"]
        )

    def test_sortGlyphNames_weightedSuffix(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a", "a.alt", "aacute", "aacute.alt", "b", "b.suf1",
                 "b.suf2", "b.suf3", "a.sc"],
                sortDescriptors=[dict(type="weightedSuffix")]
            ),
            ["a", "aacute", "b", "a.alt", "aacute.alt", "a.sc", "b.suf1",
             "b.suf2", "b.suf3"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["a", "a.alt", "aacute", "aacute.alt", "b", "b.suf1",
                 "b.suf2", "b.suf3", "a.sc"],
                sortDescriptors=[
                    dict(type="weightedSuffix", ascending=False)]
            ),
            ["b.suf3", "b.suf2", "b.suf1", "a.sc", "aacute.alt", "a.alt", "a",
             "aacute", "b", ]
        )

    def test_sortGlyphNames_ligature(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a", "a_b", "c", "c_a"],
                sortDescriptors=[dict(type="ligature")]),
            ["b", "a", "c", "a_b", "c_a"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a", "a_b", "c", "c_a"],
                sortDescriptors=[dict(type="ligature", ascending=False)]),
            ["c_a", "a_b", "c", "a", "b"]
        )

    def test_sortGlyphNames_cannedDesign(self):
        font = self.font
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a", "a_b", "c", "c_a", "a.alt", "a.suf1", "b.suf1",
                 "a.suf2", "b.suf2", "schwa", "alpha", "comma", "aacute",
                 "aacute.alt"],
                sortDescriptors=[dict(type="cannedDesign")]),
            ["a", "aacute", "b", "c", "schwa", "alpha", "comma", "a_b", "c_a",
             "a.alt", "aacute.alt", "a.suf1", "b.suf1", "a.suf2", "b.suf2"]
        )
        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["b", "a", "a_b", "c", "c_a", "a.alt", "a.suf1", "b.suf1",
                 "a.suf2", "b.suf2", "schwa", "alpha", "comma", "aacute",
                 "aacute.alt"],
                sortDescriptors=[dict(type="cannedDesign", ascending=False)]),
            ["b.suf2", "a.suf2", "b.suf1", "a.suf1", "aacute.alt", "a.alt",
             "c_a", "a_b", "comma", "alpha", "schwa", "c", "b", "aacute", "a"]
        )

    def test_sortGlyphNames_custom(self):
        font = self.font

        def sortByE(font, glyphNames, ascending=True,
                    allowsPseudoUnicodes=False):
            startsWithE = []
            doesNotStartWithE = []
            for glyphName in glyphNames:
                if glyphName.startswith("E"):
                    startsWithE.append(glyphName)
                else:
                    doesNotStartWithE.append(glyphName)
            return [startsWithE, doesNotStartWithE]

        self.assertEqual(
            font.unicodeData.sortGlyphNames(
                ["A", "B", "C", "D", "E", "Eacute", "Egrave"],
                sortDescriptors=[dict(type="custom", function=sortByE)]),
            ["E", "Eacute", "Egrave", "A", "B", "C", "D"]
        )

    def test_endSelfNotificationObservation(self):
        font = Font()
        self.assertIsNotNone(font.unicodeData.dispatcher)
        self.assertIsNotNone(font.unicodeData.font)
        self.assertIsNotNone(font.unicodeData.layerSet)
        self.assertIsNotNone(font.unicodeData.layer)

        font.unicodeData.endSelfNotificationObservation()

        self.assertIsNone(font.unicodeData.dispatcher)
        self.assertIsNone(font.unicodeData.font)
        self.assertIsNone(font.unicodeData.layerSet)
        self.assertIsNone(font.unicodeData.layer)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseObject


class Features(BaseObject):

    """
    This object contais the test represening features in the font.

    **This object posts the following notifications:**

    - Features.Changed
    - Features.BeginUndo
    - Features.EndUndo
    - Features.BeginRedo
    - Features.EndRedo
    - Features.TextChanged

    """

    changeNotificationName = "Features.Changed"
    beginUndoNotificationName = "Features.BeginUndo"
    endUndoNotificationName = "Features.EndUndo"
    beginRedoNotificationName = "Features.BeginRedo"
    endRedoNotificationName = "Features.EndRedo"
    representationFactories = {}

    def __init__(self, font=None):
        self._font = None
        if font is not None:
            self._font = weakref.ref(font)
        super(Features, self).__init__()
        self.beginSelfNotificationObservation()
        self._dirty = False
        self._text = None

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # ----
    # Text
    # ----

    def _set_text(self, value):
        oldValue = self._text
        if oldValue == value:
            return
        self._text = value
        self.postNotification("Features.TextChanged", data=dict(oldValue=oldValue, newValue=value))
        self.dirty = True

    def _get_text(self):
        return self._text

    text = property(_get_text, _set_text, doc="The raw feature text. Setting this post *Features.TextChanged* and *Features.Changed* notifications.")

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Features, self).endSelfNotificationObservation()
        self._font = None

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        from functools import partial

        getters = [('text', partial(getattr, self))]
        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        if 'text' in data:
            self.text = data['text']



from __future__ import print_function
import os
import shutil
from pkg_resources import resource_filename
import zipfile
import fs.osfs
import fs.tempfs
import fs.zipfs
import fs.copy

TESTDATA_DIR = resource_filename("defcon.test", 'testdata')


def getTestFontPath(fileName='TestFont.ufo'):
    return os.path.join(TESTDATA_DIR, fileName)


def getTestFontCopyPath(testFontPath=None):
    if testFontPath is None:
        testFontPath = getTestFontPath()
    dirName, fileName = os.path.split(testFontPath)
    fileName, ext = os.path.splitext(fileName)
    return os.path.join(dirName, fileName + 'Copy' + ext)


def makeTestFontCopy(testFontPath=None):
    if testFontPath is None:
        testFontPath = getTestFontPath()
    copyPath = getTestFontCopyPath(testFontPath)
    if os.path.isdir(testFontPath):
        shutil.copytree(testFontPath, copyPath)
    else:
        shutil.copyfile(testFontPath, copyPath)
    return copyPath


def tearDownTestFontCopy(testFontPath=None):
    if testFontPath is None:
        testFontPath = getTestFontCopyPath()
    if os.path.isdir(testFontPath):
        shutil.rmtree(testFontPath)
    else:
        os.remove(testFontPath)


def openTestFontAsFileSystem(testFontPath=None):
    if testFontPath is None:
        testFontPath = getTestFontPath()
    if zipfile.is_zipfile(testFontPath):
        parentFS = fs.tempfs.TempFS()
        with fs.zipfs.ZipFS(testFontPath, encoding="utf-8") as origFS:
            fs.copy.copy_fs(origFS, parentFS)
            rootDirs = [
                p.name for p in parentFS.scandir(u"/")
                if p.is_dir and p.name != "__MACOSX"
            ]
            fileSystem = parentFS.opendir(
                rootDirs[0], factory=fs.subfs.ClosingSubFS
            )
    else:
        fileSystem = fs.osfs.OSFS(testFontPath)
    return fileSystem


def closeTestFontAsFileSystem(fileSystem, testFontPath=None):
    if testFontPath is None:
        testFontPath = getTestFontPath()
    if not zipfile.is_zipfile(testFontPath):
        return
    rootDir = os.path.splitext(os.path.basename(testFontPath))[0] + ".ufo"
    with fs.zipfs.ZipFS(testFontPath, write=True, encoding="utf-8") as destFS:
        fs.copy.copy_fs(fileSystem, destFS.makedir(rootDir))


class NotificationTestObserver(object):

    def __init__(self, name=None):
        self.name = name
        self.stack = []

    def __repr__(self):
        return "<_TestObservable {name} {id}>".format(name=self.name, id=id(self))

    def notificationCallback(self, notification):
        print(notification.name, notification.object.name)
        self.stack.append((notification.name, notification.object.name))

    def testCallback(self, notification):
        print(notification.name, notification.data)



from __future__ import absolute_import
import os
import weakref
from fontTools.ufoLib import UFOReader, UFOLibError
from defcon.objects.base import BaseObject


class DataSet(BaseObject):

    """
    This object manages all contents of the data directory in the font.

    **This object posts the following notifications:**

    - DataSet.Changed

    """

    changeNotificationName = "DataSet.Changed"
    representationFactories = {}

    def __init__(self, font=None):
        self._font = None
        if font is not None:
            self._font = weakref.ref(font)
        super(DataSet, self).__init__()
        self.beginSelfNotificationObservation()
        self._data = {}
        self._scheduledForDeletion = {}
        self._dirty = False

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # ----------
    # File Names
    # ----------

    def _get_fileNames(self):
        return list(self._data.keys())

    def _set_fileNames(self, fileNames):
        assert not self._data
        for fileName in fileNames:
            self._data[fileName] = _dataDict()

    fileNames = property(_get_fileNames, _set_fileNames, doc="A list of all file names. This should not be set externally.")

    # -------------
    # Dict Behavior
    # -------------

    def __getitem__(self, fileName):
        if self._data[fileName]["data"] is None:
            path = self.font.path
            reader = self.font._reader
            path = "%s/%s" % ("data", fileName)
            data = reader.readBytesFromPath(path)
            onDiskModTime = reader.getFileModificationTime(path)
            self._data[fileName] = _dataDict(data=data, onDisk=True, onDiskModTime=onDiskModTime)
        return self._data[fileName]["data"]

    def __setitem__(self, fileName, data):
        assert data is not None
        onDisk = False
        onDiskModTime = None
        if fileName in self._scheduledForDeletion:
            assert fileName not in self._data
            self._data[fileName] = self._scheduledForDeletion.pop(fileName)
        if fileName in self._data:
            n = self[fileName] # force it to load so that the stamping is correct
            onDisk = self._data[fileName]["onDisk"]
            onDiskModTime = self._data[fileName]["onDiskModTime"]
            del self._data[fileName] # now remove it
        self._data[fileName] = _dataDict(data=data, dirty=True, onDisk=onDisk, onDiskModTime=onDiskModTime)
        self.dirty = True

    def __delitem__(self, fileName):
        n = self[fileName] # force it to load so that the stamping is correct]
        self._scheduledForDeletion[fileName] = dict(self._data.pop(fileName))
        self.dirty = True

    # ----
    # Save
    # ----

    def getSaveProgressBarTickCount(self, formatVersion):
        """
        Get the number of ticks that will be used by a progress bar
        in the save method. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        return 0

    def save(self, writer, saveAs=False, progressBar=None):
        """
        Save data. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        if saveAs:
            font = self.font
            if font is not None and font.path is not None and os.path.exists(font.path):
                with UFOReader(font.path, validate=False) as reader:
                    readerDataDirectoryListing = reader.getDataDirectoryListing()
                    for fileName, data in self._data.items():
                        path = "%s/%s" % ("data", fileName)
                        if data["data"] is not None or fileName not in readerDataDirectoryListing:
                            continue
                        writer.copyFromReader(reader, path, path)
        for fileName in self._scheduledForDeletion:
            # instead of trying to maintain a list of in UFO
            # vs. in memory, simply skip and move on when
            # something can't be deleted because it isn't
            # in the UFO.
            writer.removePath("%s/%s" % ("data", fileName), force=True)
        self._scheduledForDeletion.clear()
        for fileName, data in self._data.items():
            if not data["dirty"]:
                continue
            writer.writeBytesToPath("%s/%s" % ("data", fileName), data["data"])
            data["dirty"] = False
            data["onDisk"] = True
            data["onDiskModTime"] = writer.getFileModificationTime("%s/%s" % ("data", fileName))
        self.dirty = False

    # ---------------------
    # External Edit Support
    # ---------------------

    def testForExternalChanges(self, reader):
        """
        Test for external changes. This should not be called externally.
        """
        filesOnDisk = reader.getDataDirectoryListing()
        modified = []
        added = []
        deleted = []
        for fileName in set(filesOnDisk) - set(self.fileNames):
            if fileName not in self._scheduledForDeletion:
                added.append(fileName)
            elif not self._scheduledForDeletion[fileName]["onDisk"]:
                added.append(fileName)
            elif self._scheduledForDeletion[fileName]["onDiskModTime"] != reader.getFileModificationTime(
                "%s/%s" % ("data", fileName)
            ):
                added.append(fileName)
        for fileName, data in self._data.items():
            # file on disk and has been loaded
            if fileName in filesOnDisk and data["data"] is not None:
                path = "%s/%s" % ("data", fileName)
                newModTime = reader.getFileModificationTime(path)
                if newModTime != data["onDiskModTime"]:
                    newData = reader.readBytesFromPath(path)
                    if newData != data["data"]:
                        modified.append(fileName)
                continue
            # file removed
            if fileName not in filesOnDisk and data["onDisk"]:
                deleted.append(fileName)
                continue
        return modified, added, deleted

    def reloadData(self, fileNames):
        """
        Reload specified data. This should not be called externally.
        """
        for fileName in fileNames:
            self._data[fileName] = _dataDict()
            data = self[fileName]

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(DataSet, self).endSelfNotificationObservation()
        self._font = None

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        simple_get = lambda key: self[key]

        getters = []
        for k in self.fileNames:
            getters.append((k, simple_get))

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        self._data = {}
        self._scheduledForDeletion = {}
        for k in data:
            self[k] = data[k]


def _dataDict(data=None, dirty=False, onDisk=True, onDiskModTime=None):
    return dict(data=data, dirty=dirty, onDisk=onDisk, onDiskModTime=onDiskModTime)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from __future__ import absolute_import
import weakref
from fontTools.ufoLib import UFOFileStructure
from fontTools.misc.arrayTools import unionRect
from defcon.objects.base import BaseObject
from defcon.objects.glyph import Glyph
from defcon.objects.lib import Lib
from defcon.objects.uniData import UnicodeData
from defcon.objects.color import Color
from functools import partial


class Layer(BaseObject):

    """
    This object represents a layer in a :class:`LayerSet`.

    **This object posts the following notifications:**

    - Layer.Changed
    - Layer.GlyphsChanged
    - Layer.GlyphChanged
    - Layer.GlyphWillBeAdded
    - Layer.GlyphAdded
    - Layer.GlyphWillBeDeleted
    - Layer.GlyphDeleted
    - Layer.GlyphNameChanged
    - Layer.GlyphUnicodesChanged
    - Layer.NameChanged
    - Layer.ColorChanged
    - Layer.LibChanged

    The Layer object has some dict like behavior. For example, to get a glyph::

        glyph = layer["aGlyphName"]

    To iterate over all glyphs::

        for glyph in layer:

    To get the number of glyphs::

        glyphCount = len(layer)

    To find out if a font contains a particular glyph::

        exists = "aGlyphName" in layer

    To remove a glyph::

        del layer["aGlyphName"]
    """

    changeNotificationName = "Layer.Changed"
    representationFactories = {}

    def __init__(self, layerSet=None, glyphSet=None, libClass=None, unicodeDataClass=None,
                guidelineClass=None, glyphClass=None,
                glyphContourClass=None, glyphPointClass=None, glyphComponentClass=None, glyphAnchorClass=None, glyphImageClass=None):

        if layerSet is not None:
            layerSet = weakref.ref(layerSet)
        self._layerSet = layerSet
        super(Layer, self).__init__()
        self.beginSelfNotificationObservation()

        self._name = None

        if glyphClass is None:
            glyphClass = Glyph
        if libClass is None:
            libClass = Lib
        if unicodeDataClass is None:
            unicodeDataClass = UnicodeData
        self._glyphClass = glyphClass
        self._glyphContourClass = glyphContourClass
        self._glyphPointClass = glyphPointClass
        self._glyphComponentClass = glyphComponentClass
        self._glyphAnchorClass = glyphAnchorClass
        self._glyphImageClass = glyphImageClass
        self._libClass = libClass
        self._guidelineClass = guidelineClass
        self._unicodeDataClass = unicodeDataClass

        self._color = None
        self._lib = None
        self._tempLib = None
        self._unicodeData = None

        self._directory = None

        self._glyphs = {}
        self._glyphSet = glyphSet
        self._scheduledForDeletion = {}
        self._keys = set()

        self._dirty = False

        if glyphSet is not None:
            self._keys = set(self._glyphSet.keys())

    def __del__(self):
        super(Layer, self).__del__()
        self._glyphs = None
        self._lib = None
        self._unicodeData = None

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.layerSet

    def _get_font(self):
        layerSet = self.layerSet
        if layerSet is None:
            return None
        return layerSet.font

    font = property(_get_font, doc="The :class:`Font` that this layer belongs to.")

    def _get_layerSet(self):
        if self._layerSet is None:
            return None
        return self._layerSet()

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this layer belongs to.")

    # --------------
    # Glyph Creation
    # --------------

    def instantiateGlyphObject(self):
        glyph = self._glyphClass(
            layer=self,
            contourClass=self._glyphContourClass,
            pointClass=self._glyphPointClass,
            componentClass=self._glyphComponentClass,
            anchorClass=self._glyphAnchorClass,
            guidelineClass=self._guidelineClass,
            libClass=self._libClass,
            imageClass=self._glyphImageClass
        )
        return glyph

    def beginSelfGlyphNotificationObservation(self, glyph):
        glyph.addObserver(observer=self, methodName="_glyphDirtyStateChange", notification="Glyph.Changed")
        glyph.addObserver(observer=self, methodName="_glyphNameChange", notification="Glyph.NameChanged")
        glyph.addObserver(observer=self, methodName="_glyphUnicodesChange", notification="Glyph.UnicodesChanged")

    def endSelfGlyphNotificationObservation(self, glyph):
        if glyph.dispatcher is None:
            return
        glyph.removeObserver(observer=self, notification="Glyph.Changed")
        glyph.removeObserver(observer=self, notification="Glyph.NameChanged")
        glyph.removeObserver(observer=self, notification="Glyph.UnicodesChanged")
        glyph.endSelfNotificationObservation()

    def loadGlyph(self, name):
        """
        Load a glyph from the glyph set. This should not be called
        externally, but subclasses may override it for custom behavior.
        """
        if self._glyphSet is None or name not in self._glyphSet or name in self._scheduledForDeletion:
            raise KeyError("%s not in layer" % name)
        glyph = self.instantiateGlyphObject()
        glyph.disableNotifications()
        glyph._isLoading = True
        glyph.name = name
        self._stampGlyphDataState(glyph)
        self._insertGlyph(glyph)
        pointPen = glyph.getPointPen()
        self._glyphSet.readGlyph(glyphName=name, glyphObject=glyph, pointPen=pointPen)
        glyph.dirty = False
        glyph._isLoading = False
        glyph.enableNotifications()
        return glyph

    def newGlyph(self, name):
        """
        Create a new glyph with **name**. If a glyph with that
        name already exists, the existing glyph will be replaced
        with the new glyph.

        This posts *Layer.GlyphWillBeAdded*, *Layer.GlyphAdded*
        and *Layer.Changed* notifications.
        """
        self.postNotification("Layer.GlyphWillBeAdded", data=(dict(name=name)))
        if name in self and self._unicodeData is not None:
            self._unicodeData.removeGlyphData(name, self[name].unicodes)
        glyph = self.instantiateGlyphObject()
        glyph.disableNotifications()
        glyph.name = name
        self._insertGlyph(glyph)
        glyph.enableNotifications()
        self.postNotification("Layer.GlyphAdded", data=(dict(name=name)))
        self.dirty = True
        return glyph

    def insertGlyph(self, glyph, name=None):
        """
        Insert **glyph** into the layer. Optionally, the glyph
        can be renamed at the same time by providing **name**.
        If a glyph with the glyph name, or the name provided
        as **name**, already exists, the existing glyph will
        be replaced with the new glyph.

        This posts *Layer.GlyphWillBeAdded*, *Layer.GlyphAdded*
        and *Layer.Changed* notifications.
        """
        # DO NOT ACTUALLY INSERT THE GLYPH!
        # it is crucially important that the data be reconstructed
        # in its entirety so that the parent data is properly set
        # in all of the various objects.
        # FIXME: Please Explain!!
        source = glyph
        if name is None:
            name = source.name
        self.postNotification("Layer.GlyphWillBeAdded", data=(dict(name=name)))
        self.holdNotifications(note="Requested by Layer.insertGlyph.")
        dest = self.newGlyph(name)
        dest.copyDataFromGlyph(glyph)
        self.releaseHeldNotifications()
        return dest

    def _insertGlyph(self, glyph, beginObservations=True):
        name = glyph.name
        self._glyphs[name] = glyph
        if name in self._scheduledForDeletion:
            del self._scheduledForDeletion[name]
        self._keys.add(name)
        if beginObservations:
            self.beginSelfGlyphNotificationObservation(glyph)
        if glyph.unicodes and self._unicodeData is not None:
            self._unicodeData.addGlyphData(name, glyph.unicodes)

    # -------------
    # Dict Behavior
    # -------------

    def __iter__(self):
        # this is a value iterator, unlike dict()
        for name in self.keys():
            yield self[name]

    def __getitem__(self, name):
        if name not in self._glyphs:
            self.loadGlyph(name)
        return self._glyphs[name]

    def __delitem__(self, name):
        if name not in self:
            raise KeyError("%s not in layer" % name)
        self.postNotification("Layer.GlyphWillBeDeleted", data=dict(name=name))
        self._deleteGlyph(name)
        self.postNotification("Layer.GlyphDeleted", data=dict(name=name))
        self.dirty = True

    def _deleteGlyph(self, name, endObservations=True):
        if self._unicodeData is not None:
            self._unicodeData.removeGlyphData(name, self[name].unicodes)
        dataOnDiskTimeStamp = None
        dataOnDisk = None
        if name in self._glyphs:
            glyph = self._glyphs.pop(name)
            if endObservations:
                self.endSelfGlyphNotificationObservation(glyph)
            dataOnDiskTimeStamp = glyph._dataOnDiskTimeStamp
            dataOnDisk = glyph._dataOnDisk
        if name in self._keys:
            self._keys.remove(name)
        if self._glyphSet is not None and name in self._glyphSet:
            self._scheduledForDeletion[name] = dict(dataOnDiskTimeStamp=dataOnDiskTimeStamp, dataOnDisk=dataOnDisk)

    def __len__(self):
        return len(self.keys())

    def __contains__(self, name):
        return name in self.keys()

    def keys(self):
        """
        The names of all glyphs in the layer.
        """
        # this is not generated dynamically since we
        # support external editing. it must be fixed.
        if not self._scheduledForDeletion:
            return self._keys
        return self._keys - set(self._scheduledForDeletion.keys())

    # ----------
    # Attributes
    # ----------

    # name

    def _set_name(self, value):
        value = str(value)
        oldName = self._name
        if oldName != value:
            self._name = value
            data = dict(oldName=oldName, newName=value)
            self.postNotification(notification="Layer.NameChanged", data=data)
            self.dirty = True

    def _get_name(self):
        return self._name

    name = property(_get_name, _set_name, doc="The name of the layer. Setting this posts *Layer.NameChanged* and *Layer.Changed* notifications.")

    # color

    def _get_color(self):
        return self._color

    def _set_color(self, color):
        if color is None:
            newColor = None
        else:
            newColor = Color(color)
        oldColor = self._color
        if oldColor != newColor:
            self._color = newColor
            data = dict(oldColor=oldColor, newColor=newColor)
            self.postNotification(notification="Layer.ColorChanged", data=data)
            self.dirty = True

    color = property(_get_color, _set_color, doc="The layer's :class:`Color` object. When setting, the value can be a UFO color string, a sequence of (r, g, b, a) or a :class:`Color` object. Setting this posts *Layer.ColorChanged* and *Layer.Changed* notifications.")

    # -------------
    # Data Skimmers
    # -------------

    # outlines

    def _get_glyphsWithOutlines(self):
        found = []
        # scan loaded glyphs
        for glyphName, glyph in self._glyphs.items():
            if glyphName in self._scheduledForDeletion:
                continue
            if len(glyph):
                found.append(glyphName)
        # scan glyphs that have not been loaded
        if self._glyphSet is not None:
            for glyphName, fileName in self._glyphSet.contents.items():
                if glyphName in self._glyphs or glyphName in self._scheduledForDeletion:
                    continue
                glif = self._glyphSet.getGLIF(glyphName)
                containsPoints = _fetchHasOutlineData(glif)
                if containsPoints:
                    found.append(glyphName)
        return found

    glyphsWithOutlines = property(_get_glyphsWithOutlines, doc="A list of glyphs containing outlines.")

    # component references

    def _get_componentReferences(self):
        found = {}
        # scan loaded glyphs
        for glyphName, glyph in self._glyphs.items():
            if glyphName in self._scheduledForDeletion:
                continue
            if not glyph.components:
                continue
            for component in glyph.components:
                baseGlyph = component.baseGlyph
                if baseGlyph not in found:
                    found[baseGlyph] = set()
                found[baseGlyph].add(glyphName)
        # scan glyphs that have not been loaded
        if self._glyphSet is not None:
            glyphNames = set(self._glyphSet.contents.keys()) - set(self._glyphs.keys())
            for glyphName, baseList in self._glyphSet.getComponentReferences(glyphNames).items():
                for baseGlyph in baseList:
                    if baseGlyph not in found:
                        found[baseGlyph] = set()
                    found[baseGlyph].add(glyphName)
        return found

    componentReferences = property(_get_componentReferences, doc="A dict of describing the component relationships in the layer. The dictionary is of form ``{base glyph : [references]}``.")

    # image references

    def _get_imageReferences(self):
        found = {}
        # scan loaded glyphs
        for glyphName, glyph in self._glyphs.items():
            if glyphName in self._scheduledForDeletion:
                continue
            image = glyph.image
            if image is not None and image.fileName is not None:
                fileName = image.fileName
                if fileName not in found:
                    found[fileName] = []
                found[fileName].append(glyphName)
        # scan glyphs that have not been loaded
        if self._glyphSet is not None:
            glyphNames = set(self._glyphSet.contents.keys()) - set(self._glyphs.keys())
            for glyphName, fileName in self._glyphSet.getImageReferences(glyphNames).items():
                if fileName not in found:
                    found[fileName] = []
                found[fileName].append(glyphName)
        return found

    imageReferences = property(_get_imageReferences, doc="A dict of describing the image file references in the layer. The dictionary is of form ``{image file name : [references]}``.")

    # bounds

    def _get_bounds(self):
        fontRect = None
        for glyph in self:
            glyphRect = glyph.bounds
            if glyphRect is None:
                continue
            if fontRect is None:
                fontRect = glyphRect
            else:
                fontRect = unionRect(fontRect, glyphRect)
        return fontRect

    bounds = property(_get_bounds, doc="The bounds of all glyphs in the layer. This can be an expensive operation.")

    # control point bounds

    def _get_controlPointBounds(self):
        fontRect = None
        for glyph in self:
            glyphRect = glyph.controlPointBounds
            if glyphRect is None:
                continue
            if fontRect is None:
                fontRect = glyphRect
            else:
                fontRect = unionRect(fontRect, glyphRect)
        return fontRect

    controlPointBounds = property(_get_controlPointBounds, doc="The control bounds of all glyphs in the layer. This only measures the point positions, it does not measure curves. So, curves without points at the extrema will not be properly measured. This is an expensive operation.")

    # -----------
    # Sub-Objects
    # -----------

    # lib

    def instantiateLib(self):
        lib = self._libClass(
            layer=self
        )
        return lib

    def beginSelfLibNotificationObservation(self):
        self._lib.addObserver(observer=self, methodName="_libDirtyStateChange", notification="Lib.Changed")

    def endSelfLibNotificationObservation(self):
        if self._lib is None:
            return
        if self._lib.dispatcher is None:
            return
        self._lib.removeObserver(observer=self, notification="Lib.Changed")
        self._lib.endSelfNotificationObservation()

    def _get_lib(self):
        if self._lib is None:
            self._lib = self.instantiateLib()
            self.beginSelfLibNotificationObservation()
        return self._lib

    def _set_lib(self, value):
        if value is not None:
            self.lib.clear()
            self.lib.update(value)

    lib = property(_get_lib, _set_lib, doc="The layer's :class:`Lib` object.")

    # temp lib

    def _get_tempLib(self):
        if self._tempLib is None:
            self._tempLib = self.instantiateLib()
        return self._tempLib

    def _set_tempLib(self, value):
        if value is not None:
            self.tempLib.clear()
            self.tempLib.update(value)

    tempLib = property(_get_tempLib, _set_tempLib, doc="The layer's :class:`tempLib` object.")

    # unicode data

    def instantiateUnicodeData(self):
        unicodeData = self._unicodeDataClass(
            layer=self
        )
        return unicodeData

    def beginSelfUnicodeDataNotificationObservation(self):
        pass

    def endSelfUnicodeDataNotificationObservation(self):
        if self._unicodeData is None or self._unicodeData.dispatcher is None:
            return
        self._unicodeData.endSelfNotificationObservation()

    def _get_unicodeData(self):
        if self._unicodeData is None:
            cmap = {}
            for glyphName, glyph in self._glyphs.items():
                if glyphName in self._scheduledForDeletion:
                    continue
                if not glyph.unicodes:
                    continue
                for code in glyph.unicodes:
                    if code in cmap:
                        cmap[code].append(glyphName)
                    else:
                        cmap[code] = [glyphName]
            if self._glyphSet is not None:
                glyphNames = set(self._glyphSet.keys()) - set(self._glyphs.keys())
                for glyphName, unicodes in self._glyphSet.getUnicodes(glyphNames=glyphNames).items():
                    for code in unicodes:
                        if code in cmap:
                            cmap[code].append(glyphName)
                        else:
                            cmap[code] = [glyphName]

            self._unicodeData = self.instantiateUnicodeData()
            self._unicodeData.disableNotifications()
            self._unicodeData.update(cmap)
            self._unicodeData.enableNotifications()
            self.beginSelfUnicodeDataNotificationObservation()
        return self._unicodeData

    unicodeData = property(_get_unicodeData, doc="The layer's :class:`UnicodeData` object.")

    # ----
    # Save
    # ----

    def getSaveProgressBarTickCount(self, formatVersion):
        """
        Get the number of ticks that will be used by a progress bar
        in the save method. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        return 0

    def save(self, glyphSet, saveAs=False, progressBar=None):
        """
        Save the layer. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        # for a save as operation, load all the glyphs
        # and treat them as dirty in saveGlyph. this could be more
        # efficiently handled by os.copy...
        if saveAs:
            for glyph in self:
                pass
        for glyphName, glyph in sorted(self._glyphs.items()):
            self.saveGlyph(glyph, glyphSet, saveAs=saveAs)
        # remove deleted glyphs
        if not saveAs and self._scheduledForDeletion:
            for glyphName in self._scheduledForDeletion.keys():
                if glyphName in glyphSet:
                    glyphSet.deleteGlyph(glyphName)
        glyphSet.writeContents()
        self._glyphSet = glyphSet
        self._scheduledForDeletion.clear()

    def saveGlyph(self, glyph, glyphSet, saveAs=False):
        """
        Save a glyph. This method should not be called externally.
        Subclasses may override this method to implement custom saving behavior.
        """
        if glyph.dirty or saveAs:
            glyphSet.writeGlyph(glyph.name, glyph, glyph.drawPoints)
            self._stampGlyphDataState(glyph, glyphSet=glyphSet)
            glyph.dirty = False

    # ---------------------
    # External Edit Support
    # ---------------------

    def _stampGlyphDataState(self, glyph, glyphSet=None):
        if glyphSet is None:
            glyphSet = self._glyphSet
        if glyphSet is None:
            return
        glyphName = glyph.name
        if glyphName not in glyphSet.contents:
            return
        modTime = glyphSet.getGLIFModificationTime(glyphName)
        text = glyphSet.getGLIF(glyphName)
        glyph._dataOnDisk = text
        glyph._dataOnDiskTimeStamp = modTime

    def testForExternalChanges(self, reader):
        """
        Test for external changes. This should not be called externally.
        """
        self._glyphSet = reader.getGlyphSet(layerName=self.name, validateRead=self.ufoLibReadValidate)
        glyphSet = self._glyphSet
        if glyphSet is None:
            return [], [], []
        glyphSet.rebuildContents()
        # glyphs added since we started up
        addedGlyphs = []
        for glyphName in set(self._glyphSet.keys()) - self._keys:
            # not scheduled for deletion
            if glyphName not in self._scheduledForDeletion:
                addedGlyphs.append(glyphName)
            # scheduled for deletion but not
            # what was scheduled for deletion.
            # consider this a new glyph.
            elif self._scheduledForDeletion[glyphName]["dataOnDiskTimeStamp"] != glyphSet.getGLIFModificationTime(glyphName):
                if self._scheduledForDeletion[glyphName]["dataOnDisk"] != glyphSet.getGLIFModificationTime(glyphName):
                    addedGlyphs.append(glyphName)
        # glyphs deleted since we started up
        deletedGlyphs = list(self._keys - set(glyphSet.keys()))
        # glyphs modified since loading
        modifiedGlyphs = []
        for glyphName, glyph in self._glyphs.items():
            # deleted glyph. skip.
            if glyphName not in glyphSet.contents:
                continue
            modTime = glyphSet.getGLIFModificationTime(glyphName)
            # mod time mismatch
            if modTime != glyph._dataOnDiskTimeStamp:
                text = glyphSet.getGLIF(glyphName)
                # data mismatch
                if text != glyph._dataOnDisk:
                    modifiedGlyphs.append(glyphName)
        # add loaded glyphs to the keys
        for glyphName in addedGlyphs:
            # if the glyph was deleted, but now it is new,
            # unschedule it for deletion.
            if glyphName in self._scheduledForDeletion:
                del self._scheduledForDeletion[glyphName]
            self._keys.add(glyphName)
        # done. whew.
        return modifiedGlyphs, addedGlyphs, deletedGlyphs

    def reloadGlyphs(self, glyphNames):
        """
        Reload the glyphs. This should not be called externally.
        """
        for glyphName in glyphNames:
            if glyphName not in self._glyphs:
                self.loadGlyph(glyphName)
            else:
                glyph = self._glyphs[glyphName]
                glyph.destroyAllRepresentations(None)
                glyph.clear()
                pointPen = glyph.getPointPen()
                self._glyphSet.readGlyph(glyphName=glyphName, glyphObject=glyph, pointPen=pointPen)
                glyph.dirty = False
                self._stampGlyphDataState(glyph)
        data = dict(glyphNames=glyphNames)
        # post a change notification for any glyphs that
        # reference the reloaded glyphs via components.
        componentReferences = self.componentReferences
        referenceChanges = set()
        for glyphName in glyphNames:
            if glyphName not in componentReferences:
                continue
            for reference in componentReferences[glyphName]:
                if reference in glyphNames:
                    continue
                if reference not in self._glyphs:
                    continue
                if reference in referenceChanges:
                    continue
                glyph = self._glyphs[reference]
                glyph.destroyAllRepresentations(None)
                glyph.postNotification(notification=glyph.changeNotificationName)
                referenceChanges.add(reference)

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        if self.dispatcher is None:
            return
        for glyph in self._glyphs.values():
            self.endSelfGlyphNotificationObservation(glyph)
        self.endSelfLibNotificationObservation()
        self.endSelfUnicodeDataNotificationObservation()
        super(Layer, self).endSelfNotificationObservation()
        self._layerSet = None

    def _glyphDirtyStateChange(self, notification):
        self.postNotification("Layer.GlyphChanged")
        self.dirty = True

    def _libDirtyStateChange(self, notification):
        self.postNotification("Layer.LibChanged")
        self.dirty = True

    def _glyphNameChange(self, notification):
        data = notification.data
        oldName = data["oldValue"]
        newName = data["newValue"]
        glyph = self._glyphs[oldName]
        self._deleteGlyph(oldName, endObservations=False)
        if self._unicodeData is not None:
            self._unicodeData.removeGlyphData(oldName, glyph.unicodes)
        self._insertGlyph(glyph, beginObservations=False)
        self.postNotification("Layer.GlyphNameChanged", data=dict(oldValue=oldName, newValue=newName))

    def _glyphUnicodesChange(self, notification):
        glyphName = notification.object.name
        data = notification.data
        oldValues = data["oldValue"]
        newValues = data["newValue"]
        if self._unicodeData is not None:
            self._unicodeData.removeGlyphData(glyphName, oldValues)
            self._unicodeData.addGlyphData(glyphName, newValues)
        self.postNotification("Layer.GlyphUnicodesChanged", data=dict(oldValue=oldValues, newValue=newValues))

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        simple_get = partial(getattr, self)
        serialize = lambda item: item.getDataForSerialization()
        serialized_get = lambda key: serialize(simple_get(key))

        getters = (
            ('lib', serialized_get),
            ('tempLib', serialized_get),
            ('color', simple_get),
            ('glyphs', lambda _: {name: self[name].getDataForSerialization() for name in self.keys()})
        )

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        set_attr = partial(setattr, self) # key, data

        def set_glyph(name, data):
            glyph = self.instantiateGlyphObject()
            glyph.setDataFromSerialization(data)
            # there is a redundancy of of the glyph name. I decide here that
            # the single source of truth is the dict key of the layer, not
            # whatever the glyph brings with it.
            glyph.name = name
            self._insertGlyph(glyph)

        def set_glyphs(key, glyphs):
            for name in glyphs:
                set_glyph(name, glyphs[name])

        setters = (
            ('lib', set_attr),
            ('tempLib', set_attr),
            ('color', set_attr),
            ('glyphs', set_glyphs)
        )

        for key, setter in setters:
            if key not in data:
                continue
            setter(key, data[key])


# ------------
# Fast Parsers
# ------------

# this was forked from glifLib.

def _number(s):
    try:
        n = int(s)
        return n
    except ValueError:
        pass
    n = float(s)
    return n

class _DoneParsing(Exception): pass

class _BaseParser(object):

    def __init__(self):
        self._elementStack = []

    def parse(self, text):
        from xml.parsers.expat import ParserCreate
        parser = ParserCreate()
        # no attribute returns_unicode in Python3
        if hasattr(parser, "returns_unicode"):
            parser.returns_unicode = 0
        parser.StartElementHandler = self.startElementHandler
        parser.EndElementHandler = self.endElementHandler
        parser.Parse(text)

    def startElementHandler(self, name, attrs):
        self._elementStack.append(name)

    def endElementHandler(self, name):
        other = self._elementStack.pop(-1)
        assert other == name


def _fetchControlPointBoundsData(glif):
    parser = _FetchControlPointBoundsDataParser()
    try:
        parser.parse(glif)
    except _DoneParsing:
        pass
    return list(parser.points), list(parser.components)

_onCurvePointTypes = set(("move", "line", "curve", "qcurve"))
_transformationInfo = (
    ("xScale",    1),
    ("xyScale",   0),
    ("yxScale",   0),
    ("yScale",    1),
    ("xOffset",   0),
    ("yOffset",   0),
)

class _FetchControlPointBoundsDataParser(_BaseParser):

    def __init__(self):
        self.points = set()
        self.components = []
        super(_FetchControlPointBoundsDataParser, self).__init__()

    def startElementHandler(self, name, attrs):
        if name == "point" and self._elementStack and self._elementStack[-1] == "contour":
            if attrs.get("type") in _onCurvePointTypes:
                x = attrs.get("x")
                y = attrs.get("y")
                if x is not None and y is not None:
                    x = _number(x)
                    y = _number(y)
                    self.points.add((x, y))
        elif name == "component" and self._elementStack and self._elementStack[-1] == "outline":
            base = attrs.get("base")
            transformation = []
            for attr, default in _transformationInfo:
                value = attrs.get(attr)
                if value is None:
                    value = default
                else:
                    value = _number(value)
                transformation.append(value)
            self.components.append((base, transformation))
        super(_FetchControlPointBoundsDataParser, self).startElementHandler(name, attrs)

    def endElementHandler(self, name):
        if name == "outline":
            raise _DoneParsing
        super(_FetchControlPointBoundsDataParser, self).endElementHandler(name)


def _fetchHasOutlineData(glif):
    parser = _FetchHasOutlineDataParser()
    try:
        parser.parse(glif)
    except _DoneParsing:
        pass
    return parser.hasOutline

class _FetchHasOutlineDataParser(_BaseParser):

    def __init__(self):
        self.hasOutline = False
        super(_FetchHasOutlineDataParser, self).__init__()

    def startElementHandler(self, name, attrs):
        if name == "point" and self._elementStack and self._elementStack[-1] == "contour":
            segmentType = attrs.get("type")
            if segmentType not in ("move", "offcurve", None):
                self.hasOutline = True
                raise _DoneParsing
        super(_FetchHasOutlineDataParser, self).startElementHandler(name, attrs)

    def endElementHandler(self, name):
        if name == "outline":
            raise _DoneParsing
        super(_FetchHasOutlineDataParser, self).endElementHandler(name)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
